@@ -3,21 +3,18 @@ import functools
 import io
 import os
 import unittest
-import unittest.mock as mock
 import zlib
+from unittest import mock
 
 import aiohttp.multipart
+from aiohttp import helpers
+from aiohttp.hdrs import (CONTENT_DISPOSITION, CONTENT_ENCODING,
+                          CONTENT_TRANSFER_ENCODING, CONTENT_TYPE)
 from aiohttp.helpers import parse_mimetype
-from aiohttp.hdrs import (
-    CONTENT_DISPOSITION,
-    CONTENT_ENCODING,
-    CONTENT_TRANSFER_ENCODING,
-    CONTENT_TYPE
-)
-from aiohttp.multipart import (
-    parse_content_disposition,
-    content_disposition_filename
-)
+from aiohttp.multipart import (content_disposition_filename,
+                               parse_content_disposition)
+from aiohttp.streams import DEFAULT_LIMIT as stream_reader_default_limit
+from aiohttp.streams import StreamReader
 
 
 def run_in_loop(f):
@@ -48,7 +45,7 @@ class TestCase(unittest.TestCase, metaclass=MetaAioTestCase):
         self.loop.close()
 
     def future(self, obj):
-        fut = asyncio.Future(loop=self.loop)
+        fut = helpers.create_future(self.loop)
         fut.set_result(obj)
         return fut
 
@@ -175,7 +172,7 @@ class PartReaderTestCase(TestCase):
         stream = Stream(b'')
 
         def prepare(data):
-            f = asyncio.Future(loop=self.loop)
+            f = helpers.create_future(self.loop)
             f.set_result(data)
             return f
 
@@ -216,7 +213,7 @@ class PartReaderTestCase(TestCase):
         stream = Stream(b'')
 
         def prepare(data):
-            f = asyncio.Future(loop=self.loop)
+            f = helpers.create_future(self.loop)
             f.set_result(data)
             return f
 
@@ -260,14 +257,13 @@ class PartReaderTestCase(TestCase):
         self.assertEqual(b'.' * size, result)
         self.assertTrue(obj.at_eof())
 
-    def test_read_does_reads_boundary(self):
+    def test_read_does_not_read_boundary(self):
         stream = Stream(b'Hello, world!\r\n--:')
         obj = aiohttp.multipart.BodyPartReader(
             self.boundary, {}, stream)
         result = yield from obj.read()
         self.assertEqual(b'Hello, world!', result)
-        self.assertEqual(b'', (yield from stream.read()))
-        self.assertEqual([b'--:'], list(obj._unread))
+        self.assertEqual(b'--:', (yield from stream.read()))
 
     def test_multiread(self):
         obj = aiohttp.multipart.BodyPartReader(
@@ -344,6 +340,15 @@ class PartReaderTestCase(TestCase):
         self.assertEqual(b'\xd0\x9f\xd1\x80\xd0\xb8\xd0\xb2\xd0\xb5\xd1\x82,'
                          b' \xd0\xbc\xd0\xb8\xd1\x80!', result)
 
+    def test_read_with_content_transfer_encoding_binary(self):
+        obj = aiohttp.multipart.BodyPartReader(
+            self.boundary, {CONTENT_TRANSFER_ENCODING: 'binary'},
+            Stream(b'\xd0\x9f\xd1\x80\xd0\xb8\xd0\xb2\xd0\xb5\xd1\x82,'
+                   b' \xd0\xbc\xd0\xb8\xd1\x80!\r\n--:--'))
+        result = yield from obj.read(decode=True)
+        self.assertEqual(b'\xd0\x9f\xd1\x80\xd0\xb8\xd0\xb2\xd0\xb5\xd1\x82,'
+                         b' \xd0\xbc\xd0\xb8\xd1\x80!', result)
+
     def test_read_with_content_transfer_encoding_unknown(self):
         obj = aiohttp.multipart.BodyPartReader(
             self.boundary, {CONTENT_TRANSFER_ENCODING: 'unknown'},
@@ -357,6 +362,13 @@ class PartReaderTestCase(TestCase):
         result = yield from obj.text()
         self.assertEqual('Hello, world!', result)
 
+    def test_read_text_default_encoding(self):
+        obj = aiohttp.multipart.BodyPartReader(
+            self.boundary, {},
+            Stream('Привет, Мир!\r\n--:--'.encode('utf-8')))
+        result = yield from obj.text()
+        self.assertEqual('Привет, Мир!', result)
+
     def test_read_text_encoding(self):
         obj = aiohttp.multipart.BodyPartReader(
             self.boundary, {},
@@ -463,8 +475,7 @@ class PartReaderTestCase(TestCase):
             self.boundary, {}, stream)
         yield from obj.release()
         self.assertTrue(obj.at_eof())
-        self.assertEqual(b'\r\nworld!\r\n--:--', stream.content.read())
-        self.assertEqual([b'--:\r\n'], list(obj._unread))
+        self.assertEqual(b'--:\r\n\r\nworld!\r\n--:--', stream.content.read())
 
     def test_release_respects_content_length(self):
         obj = aiohttp.multipart.BodyPartReader(
@@ -480,8 +491,7 @@ class PartReaderTestCase(TestCase):
             self.boundary, {}, stream)
         yield from obj.release()
         yield from obj.release()
-        self.assertEqual(b'\r\nworld!\r\n--:--', stream.content.read())
-        self.assertEqual([b'--:\r\n'], list(obj._unread))
+        self.assertEqual(b'--:\r\n\r\nworld!\r\n--:--', stream.content.read())
 
     def test_filename(self):
         part = aiohttp.multipart.BodyPartReader(
@@ -490,6 +500,16 @@ class PartReaderTestCase(TestCase):
             None)
         self.assertEqual('foo.html', part.filename)
 
+    def test_reading_long_part(self):
+        size = 2 * stream_reader_default_limit
+        stream = StreamReader()
+        stream.feed_data(b'0' * size + b'\r\n--:--')
+        stream.feed_eof()
+        obj = aiohttp.multipart.BodyPartReader(
+            self.boundary, {}, stream)
+        data = yield from obj.read()
+        self.assertEqual(len(data), size)
+
 
 class MultipartReaderTestCase(TestCase):
 
@@ -583,6 +603,7 @@ class MultipartReaderTestCase(TestCase):
                    b'\r\n'
                    b'passed\r\n'
                    b'----:----\r\n'
+                   b'\r\n'
                    b'--:--'))
         yield from reader.release()
         self.assertTrue(reader.at_eof())
@@ -683,6 +704,24 @@ class MultipartReaderTestCase(TestCase):
             body_parts.append(read_part)
         self.assertListEqual(body_parts, [b'chunk', b'two_chunks'])
 
+    def test_reading_skips_prelude(self):
+        reader = aiohttp.multipart.MultipartReader(
+            {CONTENT_TYPE: 'multipart/related;boundary=":"'},
+            Stream(b'Multi-part data is not supported.\r\n'
+                   b'\r\n'
+                   b'--:\r\n'
+                   b'\r\n'
+                   b'test\r\n'
+                   b'--:\r\n'
+                   b'\r\n'
+                   b'passed\r\n'
+                   b'--:--'))
+        first = yield from reader.next()
+        self.assertIsInstance(first, aiohttp.multipart.BodyPartReader)
+        second = yield from reader.next()
+        self.assertTrue(first.at_eof())
+        self.assertFalse(second.at_eof())
+
 
 class BodyPartWriterTestCase(unittest.TestCase):
 
@@ -712,10 +751,10 @@ class BodyPartWriterTestCase(unittest.TestCase):
                          self.part._guess_content_type('foo'))
 
         here = os.path.dirname(__file__)
-        filename = os.path.join(here, 'software_development_in_picture.jpg')
+        filename = os.path.join(here, 'aiohttp.png')
 
         with open(filename, 'rb') as f:
-            self.assertEqual('image/jpeg',
+            self.assertEqual('image/png',
                              self.part._guess_content_type(f))
 
     def test_guess_filename(self):
@@ -809,18 +848,50 @@ class BodyPartWriterTestCase(unittest.TestCase):
         multipart = aiohttp.multipart.MultipartWriter(boundary=':')
         multipart.append('foo-bar-baz')
         multipart.append_json({'test': 'passed'})
+        multipart.append_form({'test': 'passed'})
+        multipart.append_form([('one', 1), ('two', 2)])
+        sub_multipart = aiohttp.multipart.MultipartWriter(boundary='::')
+        sub_multipart.append('nested content')
+        sub_multipart.headers['X-CUSTOM'] = 'test'
+        multipart.append(sub_multipart)
         self.assertEqual(
             [b'--:\r\n',
-             b'CONTENT-TYPE: text/plain; charset=utf-8\r\n'
-             b'CONTENT-LENGTH: 11',
+             b'Content-Type: text/plain; charset=utf-8\r\n'
+             b'Content-Length: 11',
              b'\r\n\r\n',
              b'foo-bar-baz',
              b'\r\n',
+
              b'--:\r\n',
-             b'CONTENT-TYPE: application/json',
+             b'Content-Type: application/json',
              b'\r\n\r\n',
              b'{"test": "passed"}',
              b'\r\n',
+
+             b'--:\r\n',
+             b'Content-Type: application/x-www-form-urlencoded',
+             b'\r\n\r\n',
+             b'test=passed',
+             b'\r\n',
+
+             b'--:\r\n',
+             b'Content-Type: application/x-www-form-urlencoded',
+             b'\r\n\r\n',
+             b'one=1&two=2',
+             b'\r\n',
+
+             b'--:\r\n',
+             b'Content-Type: multipart/mixed; boundary="::"\r\nX-Custom: test',
+             b'\r\n\r\n',
+             b'--::\r\n',
+             b'Content-Type: text/plain; charset=utf-8\r\n'
+             b'Content-Length: 14',
+             b'\r\n\r\n',
+             b'nested content',
+             b'\r\n',
+             b'--::--\r\n',
+             b'',
+             b'\r\n',
              b'--:--\r\n',
              b''],
             list(self.part._serialize_multipart(multipart))
@@ -837,8 +908,8 @@ class BodyPartWriterTestCase(unittest.TestCase):
         part = aiohttp.multipart.BodyPartWriter(
             'Time to Relax!', {CONTENT_ENCODING: 'gzip'})
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-ENCODING: gzip\r\n'
-                         b'CONTENT-TYPE: text/plain; charset=utf-8',
+        self.assertEqual(b'Content-Encoding: gzip\r\n'
+                         b'Content-Type: text/plain; charset=utf-8',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -853,8 +924,8 @@ class BodyPartWriterTestCase(unittest.TestCase):
         part = aiohttp.multipart.BodyPartWriter(
             'Time to Relax!', {CONTENT_ENCODING: 'deflate'})
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-ENCODING: deflate\r\n'
-                         b'CONTENT-TYPE: text/plain; charset=utf-8',
+        self.assertEqual(b'Content-Encoding: deflate\r\n'
+                         b'Content-Type: text/plain; charset=utf-8',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -867,9 +938,9 @@ class BodyPartWriterTestCase(unittest.TestCase):
         part = aiohttp.multipart.BodyPartWriter(
             thing, {CONTENT_ENCODING: 'identity'})
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-ENCODING: identity\r\n'
-                         b'CONTENT-TYPE: application/octet-stream\r\n'
-                         b'CONTENT-LENGTH: 16',
+        self.assertEqual(b'Content-Encoding: identity\r\n'
+                         b'Content-Type: application/octet-stream\r\n'
+                         b'Content-Length: 16',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -887,8 +958,8 @@ class BodyPartWriterTestCase(unittest.TestCase):
         part = aiohttp.multipart.BodyPartWriter(
             'Time to Relax!', {CONTENT_TRANSFER_ENCODING: 'base64'})
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-TRANSFER-ENCODING: base64\r\n'
-                         b'CONTENT-TYPE: text/plain; charset=utf-8',
+        self.assertEqual(b'Content-Transfer-Encoding: base64\r\n'
+                         b'Content-Type: text/plain; charset=utf-8',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -903,8 +974,8 @@ class BodyPartWriterTestCase(unittest.TestCase):
             {CONTENT_TRANSFER_ENCODING: 'base64'})
         part._chunk_size = 6
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-TRANSFER-ENCODING: base64\r\n'
-                         b'CONTENT-TYPE: application/octet-stream',
+        self.assertEqual(b'Content-Transfer-Encoding: base64\r\n'
+                         b'Content-Type: application/octet-stream',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -918,8 +989,8 @@ class BodyPartWriterTestCase(unittest.TestCase):
         part = aiohttp.multipart.BodyPartWriter(
             'Привет, мир!', {CONTENT_TRANSFER_ENCODING: 'quoted-printable'})
         stream = part.serialize()
-        self.assertEqual(b'CONTENT-TRANSFER-ENCODING: quoted-printable\r\n'
-                         b'CONTENT-TYPE: text/plain; charset=utf-8',
+        self.assertEqual(b'Content-Transfer-Encoding: quoted-printable\r\n'
+                         b'Content-Type: text/plain; charset=utf-8',
                          next(stream))
         self.assertEqual(b'\r\n\r\n', next(stream))
 
@@ -928,6 +999,21 @@ class BodyPartWriterTestCase(unittest.TestCase):
         self.assertEqual(b'\r\n', next(stream))
         self.assertIsNone(next(stream, None))
 
+    def test_serialize_with_content_transfer_encoding_binary(self):
+        part = aiohttp.multipart.BodyPartWriter(
+            'Привет, мир!'.encode('utf-8'),
+            {CONTENT_TRANSFER_ENCODING: 'binary'})
+        stream = part.serialize()
+        self.assertEqual(b'Content-Transfer-Encoding: binary\r\n'
+                         b'Content-Type: application/octet-stream',
+                         next(stream))
+        self.assertEqual(b'\r\n\r\n', next(stream))
+
+        self.assertEqual(b'\xd0\x9f\xd1\x80\xd0\xb8\xd0\xb2\xd0\xb5\xd1\x82,'
+                         b' \xd0\xbc\xd0\xb8\xd1\x80!', next(stream))
+        self.assertEqual(b'\r\n', next(stream))
+        self.assertIsNone(next(stream, None))
+
     def test_serialize_with_content_transfer_encoding_unknown(self):
         part = aiohttp.multipart.BodyPartWriter(
             'Time to Relax!', {CONTENT_TRANSFER_ENCODING: 'unknown'})
@@ -938,6 +1024,13 @@ class BodyPartWriterTestCase(unittest.TestCase):
         self.part.set_content_disposition('related', filename='foo.html')
         self.assertEqual('foo.html', self.part.filename)
 
+    def test_wrap_multipart(self):
+        writer = aiohttp.multipart.MultipartWriter(boundary=':')
+        part = aiohttp.multipart.BodyPartWriter(writer)
+        self.assertEqual(part.headers, writer.headers)
+        part.headers['X-Custom'] = 'test'
+        self.assertEqual(part.headers, writer.headers)
+
 
 class MultipartWriterTestCase(unittest.TestCase):
 
@@ -1004,6 +1097,14 @@ class MultipartWriterTestCase(unittest.TestCase):
         self.assertEqual(part.headers[CONTENT_TYPE],
                          'application/x-www-form-urlencoded')
 
+    def test_append_multipart(self):
+        subwriter = aiohttp.multipart.MultipartWriter(boundary=':')
+        subwriter.append_json({'foo': 'bar'})
+        self.writer.append(subwriter, {CONTENT_TYPE: 'test/passed'})
+        self.assertEqual(1, len(self.writer))
+        part = self.writer.parts[0]
+        self.assertEqual(part.headers[CONTENT_TYPE], 'test/passed')
+
     def test_serialize(self):
         self.assertEqual([b''], list(self.writer.serialize()))
 
@@ -1014,6 +1115,21 @@ class MultipartWriterTestCase(unittest.TestCase):
             writer.append_json({'baz': True})
         self.assertEqual(3, len(writer))
 
+    def test_append_int_not_allowed(self):
+        with self.assertRaises(TypeError):
+            with aiohttp.multipart.MultipartWriter(boundary=':') as writer:
+                writer.append(1)
+
+    def test_append_float_not_allowed(self):
+        with self.assertRaises(TypeError):
+            with aiohttp.multipart.MultipartWriter(boundary=':') as writer:
+                writer.append(1.1)
+
+    def test_append_none_not_allowed(self):
+        with self.assertRaises(TypeError):
+            with aiohttp.multipart.MultipartWriter(boundary=':') as writer:
+                writer.append(None)
+
 
 class ParseContentDispositionTestCase(unittest.TestCase):
     # http://greenbytes.de/tech/tc2231/
