@@ -3,7 +3,9 @@ from unittest import mock
 
 import pytest
 
-from aiohttp import CIMultiDict, WSMessage, WSMsgType, errors, helpers, signals
+from aiohttp import (CIMultiDict, WSMessage, WSMsgType, errors, helpers,
+                     signals, web)
+from aiohttp.log import ws_logger
 from aiohttp.test_utils import make_mocked_coro, make_mocked_request
 from aiohttp.web import HTTPBadRequest, HTTPMethodNotAllowed, WebSocketResponse
 from aiohttp.web_ws import WebSocketReady
@@ -247,53 +249,63 @@ def test_closed_after_ctor():
 
 
 @asyncio.coroutine
-def test_send_str_closed(make_request):
+def test_send_str_closed(make_request, mocker):
     req = make_request('GET', '/')
     ws = WebSocketResponse()
     yield from ws.prepare(req)
     yield from ws.close()
-    with pytest.raises(RuntimeError):
-        ws.send_str('string')
+
+    mocker.spy(ws_logger, 'warning')
+    ws.send_str('string')
+    assert ws_logger.warning.called
 
 
 @asyncio.coroutine
-def test_send_bytes_closed(make_request):
+def test_send_bytes_closed(make_request, mocker):
     req = make_request('GET', '/')
     ws = WebSocketResponse()
     yield from ws.prepare(req)
     yield from ws.close()
-    with pytest.raises(RuntimeError):
-        ws.send_bytes(b'bytes')
+
+    mocker.spy(ws_logger, 'warning')
+    ws.send_bytes(b'bytes')
+    assert ws_logger.warning.called
 
 
 @asyncio.coroutine
-def test_send_json_closed(make_request):
+def test_send_json_closed(make_request, mocker):
     req = make_request('GET', '/')
     ws = WebSocketResponse()
     yield from ws.prepare(req)
     yield from ws.close()
-    with pytest.raises(RuntimeError):
-        ws.send_json({'type': 'json'})
+
+    mocker.spy(ws_logger, 'warning')
+    ws.send_json({'type': 'json'})
+    assert ws_logger.warning.called
 
 
 @asyncio.coroutine
-def test_ping_closed(make_request):
+def test_ping_closed(make_request, mocker):
     req = make_request('GET', '/')
     ws = WebSocketResponse()
     yield from ws.prepare(req)
     yield from ws.close()
-    with pytest.raises(RuntimeError):
-        ws.ping()
+
+    mocker.spy(ws_logger, 'warning')
+    ws.ping()
+    assert ws_logger.warning.called
 
 
 @asyncio.coroutine
-def test_pong_closed(make_request):
+def test_pong_closed(make_request, mocker):
     req = make_request('GET', '/')
     ws = WebSocketResponse()
     yield from ws.prepare(req)
     yield from ws.close()
-    with pytest.raises(RuntimeError):
-        ws.pong()
+
+    mocker.spy(ws_logger, 'warning')
+    ws.pong()
+    assert ws_logger.warning.called
 
 
 @asyncio.coroutine
@@ -360,6 +372,8 @@ def test_receive_exc_in_reader(make_request, loop, reader):
     res = helpers.create_future(loop)
     res.set_exception(exc)
     reader.read = make_mocked_coro(res)
+    ws._resp_impl.transport.drain.return_value = helpers.create_future(loop)
+    ws._resp_impl.transport.drain.return_value.set_result(True)
 
     msg = yield from ws.receive()
     assert msg.type == WSMsgType.ERROR
@@ -409,7 +423,7 @@ def test_receive_client_disconnected(make_request, loop, reader):
 
     msg = yield from ws.receive()
     assert ws.closed
-    assert msg.type == WSMsgType.CLOSE
+    assert msg.type == WSMsgType.CLOSED
     assert msg.type is msg.tp
     assert msg.data is None
     assert ws.exception() is None
@@ -443,7 +457,7 @@ def test_concurrent_receive(make_request):
 
 
 @asyncio.coroutine
-def test_close_exc(make_request, reader, loop):
+def test_close_exc(make_request, reader, loop, mocker):
     req = make_request('GET', '/')
 
     ws = WebSocketResponse()
@@ -452,6 +466,8 @@ def test_close_exc(make_request, reader, loop):
     exc = ValueError()
     reader.read.return_value = helpers.create_future(loop)
     reader.read.return_value.set_exception(exc)
+    ws._resp_impl.transport.drain.return_value = helpers.create_future(loop)
+    ws._resp_impl.transport.drain.return_value.set_result(True)
 
     yield from ws.close()
     assert ws.closed
@@ -500,3 +516,8 @@ def test_can_start_ok(make_request):
     ws = WebSocketResponse(protocols=('chat',))
     with pytest.warns(DeprecationWarning):
         assert (True, 'chat') == ws.can_start(req)
+
+
+def test_msgtype_alias():
+    # deprecated since 1.0
+    assert web.MsgType is WSMsgType
